(** Module Ast *)

type memspace = Local | Shared | Global

type binop =
    Mult    | Div     | Plus    | Minus   | Mod
  | Lshift  | Rshift  | BitOr   | BitAnd  | BitXor
  | Equal   | NotEq   | LessEq  | Less    | GreaterEq | Greater
  | Map     | Reduce  | Scanl   | Scanr   | Or        | And

type unop =
    Not | Negative | BitComplement

type loom_type =
    Bool | Void | Int32 | Int64 | Uint32 | Uint64 | Float32 | Float64
  | Array   of loom_type * expr
  | Record  of (string * loom_type) list
  | Type    of string  (* to be associated with a loom_type later *)

and expr =
    IntLiteral    of int
  | FloatLiteral  of float
  | BoolLiteral   of bool
  | Convert       of expr * loom_type
  | Place         of place
  | Binop         of expr * binop * expr
  | Unop          of unop * expr
  | Call          of string * expr list
  | NullExpr

and place =
    Aref          of string * expr
  | Saref         of string * expr
  | Rref          of string * string
  | Id            of string

type stmt =
    Block   of stmt list
  | Vdecl   of memspace * string * loom_type * expr option
  | Tdecl   of string * loom_type
  | Expr    of expr
  | Return  of expr
  | Assign  of place * expr
  | If      of expr * stmt list * stmt list
  | For     of string * loom_type * expr * expr * stmt list
  | While   of expr * stmt list
  | Break
  | Continue
  | Sync
  | NullStmt  
  
type func = {
  fname    : string;
  formals  : (string * loom_type) list;
  ret_type : loom_type;
  body     : stmt list;
}

type kernel = {
  kname    : string;
  kformals : (string * loom_type) list;
  kbody    : stmt list;
}

type program = kernel option * func list

(* -- String conversion functions -- *)

let string_of_memspace =
  function 
    Local -> "local" | Global -> "global" | Shared -> "shared"

let string_of_binop = function
    Mult -> "*" | Div -> "/" | Plus -> "+" | Minus -> "-" | Mod -> "%"
  | Lshift -> "<<" | Rshift -> ">>" | BitOr -> "|" | BitAnd -> "&" | BitXor -> "^"
  | Equal -> "=" | NotEq -> "!=" | LessEq -> "<=" | Less -> "<" | GreaterEq -> ">="
  | Greater -> ">" | Map -> "//" | Reduce -> "/." | Scanl -> "/:" | Scanr -> ":/"
  | And -> "&&" | Or -> "||"

let string_of_unop = function
  Not -> "!" | Negative -> "-" | BitComplement -> "~"

let rec string_of_type =
  let str_field (s,t) = s ^ ": " ^ string_of_type t in
    function
      Bool -> "Bool" | Void -> "Void" | Int32 -> "Int32" | Int64 -> "Int64"
    | Uint32 -> "Uint32" | Uint64 -> "Uint64"
    | Float32 -> "Float32"| Float64 -> "Float64"
    | Array(t,e)      ->  string_of_type t ^ "[" ^ string_of_expr e ^ "]"
    | Record(fields)  ->  "{" ^ String.concat ", " (List.map str_field fields)
                          ^ "}"
    | Type(name)      ->  name

and string_of_expr = function
    IntLiteral(n)   ->  string_of_int n
  | FloatLiteral(x) ->  string_of_float x
  | BoolLiteral(p)  ->  if p then "true" else "false"
  | Convert(e,t)    ->  "(" ^ string_of_expr e ^ ") :: " ^ string_of_type t
  | Place(p)        ->  string_of_place p
  | Binop(e1,op,e2) ->  "(" ^ string_of_expr e1 ^ " "
                        ^ string_of_binop op ^ " "
                        ^ string_of_expr e2 ^ ")"
  | Unop(op,e)      ->  "(" ^ string_of_unop op ^ " " ^ string_of_expr e ^ ")"
  | Call(f,args)    ->  f ^ "("
                        ^ String.concat ", " (List.map string_of_expr args)
                        ^ ")"
  | NullExpr        ->  ""

and string_of_place = function
    Aref(a,e)       ->  a ^ "[" ^ string_of_expr e ^ "]"
  | Saref(a,e)      ->  a ^ "[[" ^ string_of_expr e ^ "]]"
  | Rref(v,x)       ->  v ^ "." ^ x
  | Id(s)           ->  s

(* FIXME - should add hierarchical indentation for blocks -- would need to
   pass around indentation parameter everywhere, though *)

let rec string_of_stmt = function
    Block(stmts)      ->  "{\n"
                          ^ String.concat "" (List.map string_of_stmt stmts)
                          ^ "}\n"
  | Vdecl(m,v,t,e)    ->  string_of_memspace m ^ " " ^ v ^ ": " ^ string_of_type t
                          ^ (match e with
                               Some(e') -> " <- " ^ string_of_expr e'
                             | None     -> "")
                          ^ ";\n"
  | Tdecl(name,t)     ->  "type " ^ name ^ string_of_type t ^ ";\n"
  | Expr(e)           ->  string_of_expr(e) ^ ";\n"
  | Return(e)         ->  "return " ^ string_of_expr e ^ ";\n"
  | Assign(p,e)       ->  string_of_place p ^ " <- " ^ string_of_expr e ^ ";\n"
  | If(e,ss,[])       ->  "if " ^ string_of_expr e ^ " then {\n"
                          ^ String.concat "" (List.map string_of_stmt ss)
                          ^ "}\n"
  | If(e,ss,ss')      ->  "if " ^ string_of_expr e ^ " then {\n"
                          ^ String.concat "" (List.map string_of_stmt ss)
                          ^ "}\nelse {\n"
                          ^ String.concat "" (List.map string_of_stmt ss')
                          ^ "}\n"
  | For(v,t,e1,e2,ss) ->  "for " ^ v ^ ": " ^ string_of_type t ^ " <- "
                          ^ string_of_expr e1 ^ " .. " ^ string_of_expr e2 ^ "\n"
                          ^ String.concat "" (List.map string_of_stmt ss)
                          ^ "}\n"
  | While(e,ss)       ->  "while " ^ string_of_expr e
                          ^ String.concat "" (List.map string_of_stmt ss)
                          ^ "}\n"
  | Break             ->  "break;\n"
  | Continue          ->  "continue;\n"
  | Sync              ->  "sync;\n"
  | NullStmt          ->  "\n"
;;

let string_of_formal (name, t) = 
  name ^ ": " ^ string_of_type t 

let string_of_func f =
  "func " ^ f.fname ^ " ("
  ^ String.concat ", " (List.map string_of_formal f.formals)
  ^ "): " ^ string_of_type f.ret_type ^ " {\n"
  ^ String.concat "" (List.map string_of_stmt f.body)
  ^ "}\n"

let string_of_kernel k =
  "kernel " ^ k.kname ^ " ("
  ^ String.concat ", " (List.map string_of_formal k.kformals)
  ^ ") {\n"
  ^ String.concat "" (List.map string_of_stmt k.kbody)
  ^ "}\n"

let to_string (k,fs) =
  String.concat "" (List.map string_of_func fs)
  ^ match k with
      Some(k')  ->  string_of_kernel k'
    | None      ->  "-- Warning: no kernel supplied\n"
