(** Module Compiler *)

(** PTX virtual machine types
    Pred is used for conditionals, U16 is used for thread constants *)
type ptx_t = Pred | Void | S32 | S64 | U32 | U64 | F32 | F64 | B32 | B64 | U16

(** Memory address spaces. Param is the space of actual arguments to the kernel *)
type memspace = Local | Shared | Global | Param | Reg

type ptx_line =
    Direct      of  string
  | Entry       of  string * (ptx_t * string) list
  | Func        of  ptx_t * string * (ptx_t * string) list
  | Decl        of  memspace * ptx_t * string
  | Label       of  string
  | Branch      of  string option * string  (* option for conditioned branch *)
  | Op2         of  string * ptx_t * string * string
  | Op3         of  string * ptx_t * string * string * string
  | Convert     of  ptx_t * ptx_t * string * string
  | Load        of  memspace * ptx_t * string * string * int
  | Store       of  memspace * ptx_t * string * int * string

type ptx = ptx_line list

(* FIXME - incomplete LOOM type implementation; only covers basic types and
  arrays of basic types *)
type full_type = 
    Basic   of ptx_t 
  | ArrayN  of ptx_t * int      (* Size of array is integer literal *)
  | ArrayC  of ptx_t * string   (* Size of array specified by thread const *)

(* Type for translation environment *)
type symbols = {
  parent            : symbols option
; mutable vars      : (memspace * full_type * string) list
; return_label      : string
; continue_label    : string option
; break_label       : string option
}

type count_t = {
  mutable lab : int (* Label index count *)
; mutable pred: int (* Counts for various virtual register types *)
; mutable s32 : int
; mutable s64 : int
; mutable u32 : int
; mutable u64 : int
; mutable f32 : int
; mutable f64 : int
; mutable u16 : int
}

let counts = 
  {lab=0; pred=0; s32=0; s64=0; u32=0; u64=0; f32=0; f64=0; u16=0}
;;

(** Return consecutively-numbered labels *)
let new_label () =
  let current = counts.lab in
  counts.lab <- current + 1;
  "L" ^ string_of_int current
;;

(** Return consecutively-numbered virtual register names of a PTX type *)
let new_register = function
    Pred -> let current = counts.pred in
              counts.pred <- current + 1;
              "%rp" ^ string_of_int current 
  | S32 -> let current = counts.s32 in
              counts.s32 <- current + 1;
              "%rs" ^ string_of_int current
  | S64 -> let current = counts.s64 in
              counts.s64 <- current + 1;
              "%rsl" ^ string_of_int current
  | U32 -> let current = counts.u32 in
              counts.u32 <- current + 1;
              "%ru" ^ string_of_int current
  | U64 -> let current = counts.u64 in
              counts.u64 <- current + 1;
              "%rul" ^ string_of_int current
  | F32 -> let current = counts.f32 in
              counts.f32 <- current + 1;
              "%rf" ^ string_of_int current
  | F64 -> let current = counts.f64 in
              counts.f64 <- current + 1;
              "%rfl" ^ string_of_int current
  | U16 -> let current = counts.u16 in
              counts.u16 <- current + 1;
              "%rh" ^ string_of_int current
  | _   -> failwith "*** can't make register of this type" 
;;

let sizeof = function
    U16                           -> 2
  | Pred  | S32 | U32 | F32 | B32 -> 4
          | S64 | U64 | F64 | B64 -> 8
  | Void -> failwith "*** should not try to determine size of void type"
;;

(* -- string output -- *)

let string_of_type = function
    Pred  ->  ".pred"   | Void  ->  ""
  | S32   ->  ".s32"    | S64   ->  ".s64" 
  | U32   ->  ".u32"    | U64   ->  ".u64" 
  | F32   ->  ".f32"    | F64   ->  ".f64" 
  | B32   ->  ".b32"    | B64   ->  ".b64"
  | U16   ->  ".u16"
;;

let string_of_space = function
    Local   ->  ".local"
  | Shared  ->  ".shared"
  | Global  ->  ".global"
  | Param   ->  ".param"
  | Reg     ->  ".reg"
;;

let string_of_params (params) =
  let one (t, name) = ".param " ^ string_of_type t ^ " " ^ name in
    String.concat ", " (List.map one params)
;;

let string_of_ptx_line = function
    Direct(s)           ->  "\t"  ^ s
  | Entry(name,args)    ->  "\t.entry " ^ name ^ "(" 
                                        ^ string_of_params args ^ ")"
  | Func(Void,name,args)->  "\t.func "  ^ name ^ "(" 
                                        ^ string_of_params args ^ ")"
  | Func(t,name,args)   ->  "\t.func (" ^ string_of_params [(t,name)] ^ ") "
                                        ^ name ^ "(" 
                                        ^ string_of_params args ^ ")"
  | Decl(space,t,name)  ->  "\t" ^ string_of_space space ^ " "
                                 ^ string_of_type t ^ " " 
                                 ^ name ^ ";"
  | Label(lab)          ->  lab ^ ":"
  | Branch(None,lbl)    ->  "\tbra\t" ^ lbl ^ ";"
  | Branch(Some(p),lbl) ->  "@" ^ p ^ "\tbra\t" ^ lbl ^ ";"
  | Op2(op,t,d,a)       ->  "\t"  ^ op ^ string_of_type t ^ "\t"
                              ^ d ^ ", " ^ a ^ ";"
  | Op3(op,t,d,a,b)     ->  "\t"  ^ op ^ string_of_type t ^ "\t"
                              ^ d ^ ", " ^ a ^ ", " ^ b ^ ";"
  | Convert(t1,t2,d,a)  ->  "\tcvt" ^ string_of_type t1 ^ string_of_type t2 ^ "\t"
                              ^ d ^ ", " ^ a ^ ";"
  | Load(s,t,d,a,i)     ->  "\tld"  ^ string_of_space s ^ string_of_type t ^ "\t"
                              ^ d ^ ", [" ^ a ^ "+" ^ string_of_int i ^ "];"
  | Store(s,t,d,i,a)    ->  "\tst"  ^ string_of_space s ^ string_of_type t ^ "\t"
                              ^ "[" ^ d ^ "+" ^ string_of_int i ^ "], " ^ a ^ ";"
;;

let to_string p = 
  let strs = List.map string_of_ptx_line p in
    (String.concat "\n" strs) ^ "\n"
;;

(* -- translation functions -- *)

let trans_type = function
    Sast.Bool       ->  Pred
  | Sast.Void       ->  failwith "*** cannot translate Sast.Void type."
  | Sast.Int32      ->  S32
  | Sast.Int64      ->  S64
  | Sast.Uint32     ->  U32
  | Sast.Uint64     ->  U64
  | Sast.Float32    ->  F32
  | Sast.Float64    ->  F64
  | _               ->  U32  (* pointer for other types *)
;;

let trans_type_full t =
  match t with
    Sast.Bool   | Sast.Void   | Sast.Int32    | Sast.Int64 
  | Sast.Uint32 | Sast.Uint64 | Sast.Float32  | Sast.Float64 ->
      Basic(trans_type t)
  | Sast.Array(elem_t, Sast.IntLiteral(n)) ->
      ArrayN(trans_type elem_t, n)
  | Sast.Array(elem_t, Sast.Place(Sast.Id(Sast.Const(s, _)))) ->
      ArrayC(trans_type elem_t, s)
  | _ -> failwith "*** Not all compound types implemented."
;;

let trans_space = function
  Sast.Local -> Local   | Sast.Global -> Global 
| Sast.Shared -> Shared | Sast.Param -> Param
;;

let trans_special = function
  "T"         -> "%tid"
| "THREADS"   -> "%threads" (* FIXME -- need to think more about how these work *)
| "B"         -> "%ctaid"
| "BLOCKS"    -> "%blocks"
| "BLOCKSIZE" -> "%blocksize"
| _           -> failwith "*** unknown special constant."
;;

(** Translate expression, returning a register, instruction-list pair *)
(* Need to pass in env for array, identifier lookups *)
let rec trans_expr env (e, t) =
  let t' = trans_type t in
  let rname = new_register t' in  (* create a virtual register for each node *)
    match e with
      Sast.IntLiteral(n)    ->
        rname, [ Op2("mov", t', rname, string_of_int n) ]

    | Sast.FloatLiteral(f)  ->
        rname, [ Op2("mov", t', rname, string_of_float f) ]

    | Sast.BoolLiteral(b)   ->
        rname, [ Op2("mov", t', rname, if b then "1" else "0") ]

    | Sast.Place(p)         -> trans_place env p t

    | Sast.Binop(et1, op, et2)  ->
        let reg1, ptx1 = trans_expr env et1 in
        let reg2, ptx2 = trans_expr env et2 in
        let ptx_op =
          ptx1 
        @ ptx2
        @ begin
            match op with
              Sast.Mult   ->  [ Op3("mul", t' , rname, reg1, reg2) ]
            | Sast.Div    ->  [ Op3("div", t' , rname, reg1, reg2) ]
            | Sast.Plus   ->  [ Op3("add", t' , rname, reg1, reg2) ]
            | Sast.Minus  ->  [ Op3("sub", t' , rname, reg1, reg2) ]                
            | _ -> [ Direct("// BinOp placeholder") ]
          end
        in
          rname, ptx_op

    | Sast.Unop(op,et)       ->
        let reg1, ptx1 = trans_expr env et in
        let ptx_op =
          ptx1
        @ begin
            match op with
            | Sast.Negative       ->
                let byte_t = if (sizeof t') = 32 then B32 else B64 in
                let reg1' = new_register t' in
                  [ Op2("neg", byte_t , rname, reg1') ]
            | Sast.Not
            | Sast.BitComplement  ->
                let byte_t = if (sizeof t') = 32 then B32 else B64 in
                let reg1' = new_register t' in
                  [ Op2("not", byte_t , rname, reg1') ]
          end
        in
          rname, ptx_op
    
    | Sast.Convert((e, dest_t) as et, source_t)    ->
        (* FIXME: only works for basic types *)
        let reg1, ptx1 = trans_expr env et in
        let dest_t' = trans_type dest_t in
        let source_t' = trans_type source_t in
        let ptx_op =
          ptx1
        @ [ Convert(source_t', dest_t', rname, reg1) ]
        in
          rname, ptx_op
    
    | Sast.Call(f,args)     -> 
        rname, [ Direct("// Call placeholder") ]
    
    | Sast.NullExpr  -> rname, []

(** Translate a 'place' (lvalue), returning a register holding the address and
    a list of instructions used to calculate the address *)
and trans_place env p t = 
    "r_debug", [ Direct("// Place placeholder") ]
;;

let rec trans_stmt env stmt =
  match stmt with
    Sast.Block(stmts)   ->
      (* PTX has scope blocks built-in *)
        [ Direct("{") ]
      @ List.concat (List.map (trans_stmt env) stmts)
      @ [ Direct("}") ]

  | Sast.Vdecl(space, (name, t), init_opt)  ->  begin
      let space' = trans_space space in
      let full_t = trans_type_full t in
      let name', t' =
        match full_t with
          Basic(typ)      ->  (name, typ)
        | ArrayN(typ, n)  ->  (name ^ "[" ^ string_of_int n ^ "]", typ)
        (* FIXME - need to implement ArrayC types *)
        | _               ->  failwith "*** cannot declare compound type"
      in
        [ Decl(space', t', name') ]
    end

  | Sast.Tdecl(name, t) ->
      (* Defined types are interpreted by the typechecker, so nothing to do here *)
        []

  | Sast.Expr(et)     ->    (* throw away register holding value *)
        let rname, lst = trans_expr env et in
          lst
    
  | Sast.Return(e, t)   ->
        [ Branch(None, env.return_label) ]

  | Sast.Assign(p, e)   -> 
      (* let place_r, place_t = trans_place p *)
        [ Direct("// Assign placeholder") ]

  | Sast.If((e, t), stmts1, stmts2)     -> 
        [ Direct("// If placeholder") ]
    
  | Sast.For((name, t), e1, e2, stmts)  -> 
        [ Direct("// For placeholder") ]
    
  | Sast.While(et, stmts)        -> 
    (*  General idea:
        label1
          if (test_code) branch label2
          branch label3
        label2:
          recursively call trans_stmt on the body block with a translation
              environment having break_label = label3 and continue_label = label1
          branch label1
        label3:
      *)
        [ Direct("// For placeholder") ]
    
  | Sast.Break          -> 
        [ Direct("// Break placeholder") ]
    
  | Sast.Continue       -> 
        [ Direct("// Continue placeholder") ]
    
  | Sast.Sync           -> 
        [ Direct("bar.sync;") ]
    
  | Sast.NullStmt       -> []
;;

let trans_params params =
  let trans_one (name, t) = (trans_type t, name) in
    (List.map trans_one params) @ [(U32, "grid")]
;;

let trans_params_full params =
  let trans_one (name, t) = (Param, trans_type_full t, name) in
    (List.map trans_one params) @ [(Param, ArrayN(U32,10), "grid")]
;;

let new_environment vs =
  { parent          = None
  ; vars            = vs
  ; return_label    = new_label()
  ; continue_label  = None
  ; break_label     = None }
;;

let trans_kernel k =
  let kname = k.Sast.kname in
  let basic_params = trans_params k.Sast.kformals in
  let full_params = trans_params_full k.Sast.kformals in
  let env = new_environment full_params in
    [ Entry(kname, basic_params)
    ; Direct("{")
    ; Decl(Reg, Pred, "%rp<99>" )   (* Declare types of virtual registers *)
    ; Decl(Reg, S32,  "%rs<99>" ) 
    ; Decl(Reg, S64,  "%rsl<99>") 
    ; Decl(Reg, U32,  "%ru<99>" ) 
    ; Decl(Reg, U64,  "%rul<99>") 
    ; Decl(Reg, F32,  "%rf<99>" ) 
    ; Decl(Reg, F64,  "%rfl<99>") 
    ; Decl(Reg, B32,  "%rb<99>" ) 
    ; Decl(Reg, U16,  "%rh<99>" ) 
    ; Decl(Reg, U32,  "%threads")   (* Set up thread constants *)
    ; Decl(Reg, U32,  "%blocks" ) 
    ; Decl(Reg, U32,  "%blocksize") 
    ; Load(Param, U32, "%threads",    "grid", 0)
    ; Load(Param, U32, "%blocks",     "grid", 4)
    ; Load(Param, U32, "%blocksize",  "grid", 8) ]
  @ trans_stmt env (Sast.Block(k.Sast.kbody))
  @ [ Label(env.return_label)
    ; Direct("exit;")
    ; Direct("}")
    ]
;;

let trans_func f = [ Direct("// FUNC " ^ f.Sast.fname ^ " placeholder") ]
;;

let translate (k, fs) =
  [ Direct(".version 1.4")
  ; Direct(".target sm_10, map_f64_to_f32") ]
  @ List.concat (List.map trans_func fs)
  @ (trans_kernel k) 
;;
