(** loom.ml -- top-level command-line interface *)

let version = "LOOM version 0.1"

type filetype = Source | Pre | Parsed | Checked | Ptx

(** Attempt to determine a file's type based on its extension *)
let infer_filetype filename =
  let parts = Str.split (Str.regexp "[.]") filename in
  let extension = if List.length parts > 1 then
                    List.hd (List.rev parts)
                  else failwith "No file extension -- can't determine type."
    in
    match extension with
      "lm"  -> Source
    | "lmp" -> Pre
    | "lma" -> Parsed
    | "lms" -> Checked
    | "ptx" -> Ptx
    | _     -> failwith ("Unknown extension: " ^ extension)

(** Print string corresponding to file type *)
let string_of_filetype = function
  Source  -> "LOOM source code"
| Pre     -> "preprocessed LOOM code"
| Parsed  -> "parsed LOOM code"
| Checked -> "semantically analyzed, typechecked LOOM code"
| Ptx     -> "PTX code"

(* Parse arguments *)

let usage = version ^ "\nusage: " ^ Sys.argv.(0)
            ^ " [-p | -a | -s | -c] [-in <input>.<ext>] [-out <output>.<ext>]"

(* Default values *)
let in_fname     = ref ""   (* stdin *)
let out_fname    = ref ""   (* stdout *)
let in_filetype  = ref Source
let out_filetype = ref Ptx

let arguments = [ 
  ("-p",   Arg.Unit (fun () -> out_filetype := Pre),     ": preprocess input"   )
; ("-a",   Arg.Unit (fun () -> out_filetype := Parsed),  ": parse input"        )
; ("-s",   Arg.Unit (fun () -> out_filetype := Checked), ": typecheck input"    )
; ("-c",   Arg.Unit (fun () -> out_filetype := Ptx),     ": compile to PTX code")
; ("-in",  Arg.String (fun s -> in_fname  := s),         ": set input filename" )
; ("-out", Arg.String (fun s -> out_fname := s),         ": set output filename")
; ("-version",
        Arg.Unit (fun () -> print_endline version; exit 0),
        ": print version string and exit")
]

(* main *)

let _ =
  
  let onerr = fun arg -> raise (Arg.Bad ("Unknown argument: " ^ arg)) in
    Arg.parse arguments onerr usage;

  if !in_fname <> ""  then in_filetype  := infer_filetype !in_fname;
  if !out_fname <> "" then out_filetype := infer_filetype !out_fname;
  
  (* take appropriate action depending on the input and output filetypes given *)
  match (!in_filetype, !out_filetype) with
    Source, Pre     ->  let lines = Utx.read_lines !in_fname in
                          Utx.write_string !out_fname (Preprocessor.embrace lines)

  | Source, Parsed  ->  let lines   = Utx.read_lines !in_fname in
                        let pre     = Preprocessor.embrace lines in
                        let lexbuf  = Lexing.from_string pre in
                        let parsed  = Parser.program Scanner.token lexbuf in
                          Utx.write_string !out_fname (Ast.to_string parsed)

  | Source, Checked ->  let lines   = Utx.read_lines !in_fname in
                        let pre     = Preprocessor.embrace lines in
                        let lexbuf  = Lexing.from_string pre in
                        let parsed  = Parser.program Scanner.token lexbuf in
                        let checked = Sast.analyze parsed in
                          Utx.write_string !out_fname (Sast.to_string checked)

  | Source, Ptx     ->  let lines   = Utx.read_lines !in_fname in
                        let pre     = Preprocessor.embrace lines in
                        let lexbuf  = Lexing.from_string pre in
                        let parsed  = Parser.program Scanner.token lexbuf in
                        let checked = Sast.analyze parsed in
                        let ptx     = Compiler.translate checked in
                          Utx.write_string !out_fname (Compiler.to_string ptx)

  | Pre, Parsed     ->  let pre     = Utx.read_string !in_fname in
                        let lexbuf  = Lexing.from_string pre in
                        let parsed  = Parser.program Scanner.token lexbuf in
                          Utx.write_string !out_fname (Ast.to_string parsed)

  | Pre, Checked    ->  let pre     = Utx.read_string !in_fname in
                        let lexbuf  = Lexing.from_string pre in
                        let parsed  = Parser.program Scanner.token lexbuf in
                        let checked = Sast.analyze parsed in
                          Utx.write_string !out_fname (Sast.to_string checked)

  | Pre, Ptx        ->  let pre     = Utx.read_string !in_fname in
                        let lexbuf  = Lexing.from_string pre in
                        let parsed  = Parser.program Scanner.token lexbuf in
                        let checked = Sast.analyze parsed in
                        let ptx     = Compiler.translate checked in
                          Utx.write_string !out_fname (Compiler.to_string ptx)

  | _               ->  failwith ("*** Don't know how to translate " 
                                ^ string_of_filetype !in_filetype ^ "into"
                                ^ string_of_filetype !out_filetype ^ ".\n")
