%{ open Ast %}

%token LPAREN RPAREN LDOUBLE RDOUBLE LSQUARE RSQUARE LCURLY RCURLY
%token SEMI COMMA DOT TWODOTS THREEDOTS COLON TYPE
%token ASSIGN MAP REDUCE SCANL SCANR
%token OR AND EQ NE LE GE LT GT
%token PLUS MINUS VBAR CARET
%token STAR SLASH MOD LSHIFT RSHIFT AMP TILDE NOT CONVERT
%token ASSIGN LSHIFT RSHIFT STAR SLASH MOD PLUS MINUS VBAR CARET
%token FOR WHILE BREAK CONTINUE IF THEN ELSE PASS
%token FUNC KERNEL RETURN GLOBAL LOCAL SHARED SYNC
%token BOOL VOID INT32 INT64 UINT32 UINT64 FLOAT32 FLOAT64
%token TRUE FALSE EOF
%token <int> INTLITERAL
%token <float> FLOATLITERAL
%token <string> ID

%nonassoc NOELSE
%nonassoc ELSE
%nonassoc UMINUS
%right ASSIGN
%right MAP REDUCE SCANL SCANR
%left OR
%left AND
%left EQ NE LE GE LT GT
%left PLUS MINUS VBAR CARET
%left STAR SLASH MOD LSHIFT RSHIFT AMP
%nonassoc TILDE NOT
%left CONVERT

%start program
%type <Ast.program> program

%%

program:
    /* empty program */   { None, [] }
  | program kernel        { Some($2), snd $1 }
  | program func          { fst $1, ($2 :: snd $1) }

kernel:  KERNEL ID LPAREN args_opt RPAREN block
    { { kname     = $2;
        kformals  = $4;
        kbody     = $6 } }

func:  FUNC ID LPAREN args_opt RPAREN COLON ltype block
    { { fname     = $2;
        formals   = $4;
        ret_type  = $7;
        body      = $8 } }

args_opt:
    /* nothing */     { [] }
  | args_list { List.rev $1 }

args_list:
    ID COLON ltype                 { [($1, $3)] }
  | args_list COMMA ID COLON ltype { ($3, $5) :: $1 }

block:  LCURLY stmt_list RCURLY   { List.rev $2 }

stmt_list:
    /* nothing */     { [] }
  | stmt_list stmt    { $2 :: $1 }

stmt:
    block                                    { Block($1) }
  | expr SEMI                                { Expr($1) }
  | memspace ID COLON ltype ASSIGN expr SEMI { Vdecl($1, $2, $4, Some($6)) }
  | memspace ID COLON ltype SEMI             { Vdecl($1, $2, $4, None) }
  | TYPE ID ltype SEMI                       { Tdecl($2, $3) }
  | place ASSIGN expr SEMI                   { Assign($1, $3) }
  | IF expr THEN block                       { If($2, $4, []) }
  | IF expr THEN block ELSE block            { If($2, $4, $6) }
  | FOR ID COLON ltype ASSIGN expr TWODOTS expr block
              { For($2, $4, $6, $8, $9) }
  | FOR ID COLON ltype ASSIGN expr THREEDOTS expr block
              { For($2, $4, $6, Binop(IntLiteral(1),Plus,$8), $9) }
  | WHILE expr block                         { While ($2, $3) }
  | BREAK SEMI                               { Break }
  | CONTINUE SEMI                            { Continue }
  | RETURN expr SEMI                         { Return($2) }
  | RETURN SEMI                              { Return(NullExpr) }
  | SYNC SEMI                                { Sync }
  | PASS SEMI                                { NullStmt }

place:
    ID LSQUARE expr RSQUARE       { Aref($1, $3) }
  | ID LDOUBLE expr RDOUBLE       { Saref($1, $3) }
  | ID DOT ID                     { Rref($1, $3) }
  | ID                            { Id($1) }

expr:
    INTLITERAL                    { IntLiteral($1) }
  | FLOATLITERAL                  { FloatLiteral($1) }
  | TRUE                          { BoolLiteral(true) }
  | FALSE                         { BoolLiteral(false) }
  | LPAREN expr RPAREN            { $2 }
  | place                         { Place($1) }
  | expr CONVERT ltype            { Convert($1, $3) }
  | expr STAR expr                { Binop($1, Mult, $3) }
  | expr SLASH expr               { Binop($1, Div, $3) }
  | expr PLUS expr                { Binop($1, Plus, $3) }
  | expr MINUS expr               { Binop($1, Minus, $3) }
  | expr MOD expr                 { Binop($1, Mod, $3) }
  | expr LSHIFT expr              { Binop($1, Lshift, $3) }
  | expr RSHIFT expr              { Binop($1, Rshift, $3) }
  | expr VBAR expr                { Binop($1, BitOr, $3) }
  | expr AMP expr                 { Binop($1, BitAnd, $3) }
  | expr CARET expr               { Binop($1, BitXor, $3) }
  | expr EQ expr                  { Binop($1, Equal, $3) }
  | expr NE expr                  { Binop($1, NotEq, $3) }
  | expr LE expr                  { Binop($1, LessEq, $3) }
  | expr LT expr                  { Binop($1, Less, $3) }
  | expr GE expr                  { Binop($1, GreaterEq, $3) }
  | expr GT expr                  { Binop($1, Greater, $3) }
  | expr MAP expr                 { Binop($1, Map, $3) }
  | expr REDUCE expr              { Binop($1, Reduce, $3) }
  | expr SCANR expr               { Binop($1, Scanr, $3) }
  | expr SCANL expr               { Binop($1, Scanl, $3) }
  | NOT expr                      { Unop(Not, $2) }
  | MINUS expr %prec UMINUS       { Unop(Negative, $2) }
  | TILDE expr                    { Unop(BitComplement, $2) }
  | ID LPAREN actuals_opt RPAREN  { Call($1, $3) }

actuals_opt:
    /* nothing */   { [] }
  | actuals_list    { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }

memspace:
    LOCAL   { Local }
  | GLOBAL  { Global }
  | SHARED  { Shared }

ltype:
    BOOL    { Bool }
  | VOID    { Void }
  | INT32   { Int32 }
  | INT64   { Int64 }
  | UINT32  { Uint32 }
  | UINT64  { Uint64 }
  | FLOAT32 { Float32 }
  | FLOAT64 { Float64 }
  | ltype LSQUARE expr RSQUARE  { Array($1, $3) }
  | LCURLY fields_opt RCURLY    { Record($2) }
  | ID                          { Type($1) }

fields_opt:
    /* nothing */     { [] }
  | fields_list { List.rev $1 }

fields_list:
    ID COLON ltype                   { [($1, $3)] }
  | fields_list COMMA ID COLON ltype { ($3, $5) :: $1 }
