(** preprocessor.ml
    Convert a source file from pseudocode style, where block structure is
    indicated by indentation level, to braces-and-semicolon style suitable
    for input to the compiler front-end. Also strip out comments and blank lines.
    NB - inserts a (harmless) empty statement on the last line. *)

(** Count the number of spaces before the first non-space character in str. 
    (NB - Caml doesn't seem to encourage functional-style string manipulation,
    so fall back on refs and while-loops.) *)
let leading_spaces str =
  let count = ref 0 in
    while str.[!count] = ' ' do
      count := !count + 1
    done;
    !count;;

(** Strip off trailing spaces from str. *)
let right_strip str =
  let i = ref (String.length str) in
    while !i > 0 && str.[!i - 1] = ' ' do
      i := !i - 1
    done;
    String.sub str 0 !i

(** Given a list of indentation levels (typically just been popped from the 
    indentation stack), terminate the preceding statement with a semicolon
    and close out open blocks at each level specified. *)
let close_braces indents =
  let braces = 
    List.fold_left (fun str n -> str ^ "\n" ^ String.make n ' ' ^ "}") "" indents in
  ";" ^ braces ^ "\n";;

(** Given a state object representing prior lines, add braces and semicolons
    to the next line. *)
let embrace_line (stack, acc) line =
  let line' =                       (* line' is line sans comments *)
    match Str.bounded_split (Str.regexp "--") line 2 with
      str :: _  -> right_strip str
    | _         -> ""
  in 
  if line' = "" then
    (stack, acc)    (* ignore blank lines - pass through input state *)
  else 
    let indent = leading_spaces line' in
      match stack with
        []                        -> ([0], line' :: acc)
      | (s :: _) when indent = s  -> (stack, (";\n" ^ line') :: acc)
      | (s :: _) when indent > s  -> (indent :: stack, (" {\n" ^ line') :: acc)
      | (s :: rest)               ->
        let (to_pop, reduced) = List.partition (fun x -> x >= indent) rest in
        let closes = close_braces to_pop in
          (indent :: reduced, (closes ^ line') :: acc);;

let embrace lines =
  let lines' = lines @ ["X\n"] in  (* HACK: add a placeholder at the end *)
  let (_, backward_result) = List.fold_left embrace_line ([], []) lines' in
  let s = String.concat "" (List.rev backward_result) in
    String.sub s 0 ((String.length s) - 2)  (* HACK: delete the placeholder *)
