(** Module Sast -- semantic analyzer / typechecker *)

type memspace = Local | Shared | Global | Param

type binop =
    Mult    | Div     | Plus    | Minus   | Mod
  | Lshift  | Rshift  | BitOr   | BitAnd  | BitXor
  | Equal   | NotEq   | LessEq  | Less    | GreaterEq | Greater
  | Map     | Reduce  | Scanl   | Scanr   | And       | Or

type unop =
    Not | Negative | BitComplement

type loom_type =
    Bool | Void | Int32 | Int64 | Uint32 | Uint64 | Float32 | Float64
  | Array     of loom_type * expr
  | Record    of (string * loom_type) list
  | Function  of (loom_type list) * loom_type

and expr =
    IntLiteral    of int
  | FloatLiteral  of float
  | BoolLiteral   of bool
  | Convert       of expr_t * loom_type  (* in Sast, holds the SOURCE type *)
  | Place         of place
  | Binop         of expr_t * binop * expr_t
  | Unop          of unop * expr_t
  | Call          of fdecl * expr_t list
  | NullExpr

and place =
  | Aref          of vdecl * expr_t
  | Saref         of vdecl * expr_t
  | Rref          of vdecl * string
  | Id            of decl

and expr_t = expr * loom_type
and place_t = place * loom_type

and tdecl = string * loom_type    (* type name, type *)
and vdecl = string * loom_type    (* var name, type *)
and fdecl = string * (loom_type list) * loom_type  (* name, args, return type *)
and decl =    Var   of vdecl 
            | Func  of fdecl 
            | Const of vdecl  (* used from thread constants *)

type stmt =
    Block   of stmt list
  | Vdecl   of memspace * vdecl * expr_t option
  | Tdecl   of tdecl
  | Expr    of expr_t
  | Return  of expr_t
  | Assign  of place_t * expr_t
  | If      of expr_t * stmt list * stmt list
  | For     of vdecl * expr_t * expr_t * stmt list
  | While   of expr_t * stmt list
  | Break
  | Continue
  | Sync
  | NullStmt
  
type func = {
  fname     : string;
  formals   : vdecl list;
  ret_type  : loom_type;
  body      : stmt list;
}

type kernel = {
  kname    : string;
  kformals : vdecl list;
  kbody    : stmt list;
}

type program = kernel * func list

let thread_constants = 
    [     "T" ;   "THREADS" ;     "B" ;   "BLOCKS"  ;   "BLOCKSIZE"
    ;   "X_T" ; "X_THREADS" ;   "X_B" ; "X_BLOCKS"  ; "X_BLOCKSIZE"
    ;   "Y_T" ; "Y_THREADS" ;   "Y_B" ; "Y_BLOCKS"  ; "Y_BLOCKSIZE" ]

(* -- environment types and utility functions -- *)

type symbols = {
  parent          : symbols option
; mutable vars    : (vdecl * memspace) list
; mutable types   : tdecl list
; mutable funcs   : fdecl list
}

type environment = {
  func_name       : string
; return_type     : loom_type
; loop_level      : int
; scope           : symbols
}

(** Look up a variable name in a symbol table, and return (vdecl, memspace) if 
    found.  Otherwise, follow links upward in the scope chain.  If the variable 
    name does not appear anywhere in the chain, raise an exception. *)
let rec lookup_var_full scope vname =
  try
    List.find (fun ((s, _),_) -> s = vname) scope.vars
  with Not_found ->
    match scope.parent with
      Some(p) -> lookup_var_full p vname
    | None    -> raise Not_found
;;

(** Look up a variable name in a symbol table, and return vdecl if found. *)
let lookup_var scope vname =
  let (vd, _) = lookup_var_full scope vname in
    vd
;;

(** Look up a function name, and return fdecl if found. *)
(* FIXME: because LOOM functions have global scope, this will always go to the top
   level -- could be optimized by keeping fdecls in a separate global record *)
let rec lookup_func scope fname =
  try
    List.find (fun (s, _, _) -> s = fname) scope.funcs
  with Not_found ->
    match scope.parent with
      Some(p) -> lookup_func p fname
    | None    -> raise Not_found
;;

(** Look up a type name in scope, and return tdecl if found. *)
let rec lookup_type scope tname =
  try
    List.find (fun (s, _) -> s = tname) scope.types
  with Not_found ->
    match scope.parent with
      Some(p) -> lookup_type p tname
    | None    -> raise Not_found
;;

(** Create a string containing a summary of the scope (for error messages) *)
let rec string_of_scope scope =
  let vnames  = List.map (fun ((name,_),_)  -> name) scope.vars   in
  let tnames  = List.map (fun (name,_)      -> name) scope.types  in
  let fnames  = List.map (fun (name,_,_)    -> name) scope.funcs  in
  let parent_info = 
    match scope.parent with
      Some(p) -> "\nPARENT " ^ (string_of_scope p)
    | None    -> "(EARLIEST ANCESTOR)\n"
  in
    "SCOPE:"
    ^ "\n  vars:  " ^ String.concat ", " vnames
    ^ "\n  types: " ^ String.concat ", " tnames
    ^ "\n  funcs: " ^ String.concat ", " fnames
    ^ "\n"
    ^ parent_info
;;

(** Convert a LOOM type int a string (for error reporting) *)
let rec string_of_type =
  let str_field (s,t) = s ^ ": " ^ string_of_type t in
    function
      Bool -> "Bool" | Void -> "Void" | Int32 -> "Int32" | Int64 -> "Int64"
    | Uint32 -> "Uint32" | Uint64 -> "Uint64"
    | Float32 -> "Float32"| Float64 -> "Float64"
    | Array(t,_)      ->  string_of_type t ^ "[...]"
    | Record(fields)  ->  "{" ^ String.concat ", " (List.map str_field fields)
                          ^ "}"
    | Function(args_t,ret_t) -> 
        "Function (" 
        ^ (String.concat ", " (List.map string_of_type args_t))
        ^ "): " ^ string_of_type ret_t
;;

(* -- semantic analysis -- *)

(* FIXME - probably better to just share Ast.binop and Ast.unop types *)
let check_binop = function
    Ast.Mult    -> Mult     | Ast.Div   -> Div    | Ast.Plus -> Plus 
  | Ast.Minus   -> Minus    | Ast.Mod   -> Mod    | Ast.Lshift -> Lshift 
  | Ast.Rshift  -> Rshift   | Ast.BitOr -> BitOr  | Ast.BitAnd -> BitAnd 
  | Ast.BitXor  -> BitXor   | Ast.Equal -> Equal  | Ast.NotEq -> NotEq 
  | Ast.LessEq  -> LessEq   | Ast.Less  -> Less   | Ast.GreaterEq -> GreaterEq 
  | Ast.Greater -> Greater  | Ast.Map   -> Map    | Ast.Reduce -> Reduce 
  | Ast.Scanl   -> Scanl    | Ast.Scanr -> Scanr  
  | Ast.Or      -> Or       | Ast.And   -> And
;;

let check_unop = function
  Ast.Not -> Not | Ast.Negative -> Negative | Ast.BitComplement -> BitComplement
;;

let check_memspace = function
  Ast.Global -> Global | Ast.Local -> Local | Ast.Shared -> Shared
;;

let rec check_loom_type env = function
    Ast.Bool      -> Bool       | Ast.Void      -> Void
  | Ast.Int32     -> Int32      | Ast.Int64     -> Int64
  | Ast.Uint32    -> Uint32     | Ast.Uint64    -> Uint64
  | Ast.Float32   -> Float32    | Ast.Float64   -> Float64

  | Ast.Array(t,e)      ->
      let t'      = check_loom_type env t in
      let (e',_)  = check_expr env e in
        begin
          match e' with
            IntLiteral (_) | Place(Id(Const(_))) -> Array(t',e')
          | _   -> failwith ("\n*** Array type declaration parameter must be"
                          ^ " either an integer constant or a thread constant.")
        end

  | Ast.Record(ts)      ->
      let r = List.map (fun (s,t) -> (s, check_loom_type env t)) ts in
        (* let fieldnames, _ = List.split r in *)
        (* FIXME - check no duplicate field names *)
        Record(r)

  | Ast.Type(s)         ->
      let tdecl = try
        lookup_type env.scope s
      with Not_found    ->
        failwith ("\n*** Unknown type " ^ s)
      in
        let (_, t) = tdecl in t

(** Given an Ast expression, recursively build up a typechecked structure. *)
(* Nomenclature: "primed" variables are in Sast (e.g., e', t') *)
and check_expr env = function
    Ast.IntLiteral(n)   ->  IntLiteral(n), Int32
  | Ast.FloatLiteral(x) ->  FloatLiteral(x), Float32
  | Ast.BoolLiteral(p)  ->  BoolLiteral(p), Bool
  
  | Ast.Convert(e, dest_t)  ->            (* in Ast, Convert takes dest type *)
      let e', source_t' = check_expr env e in
      let dest_t' = check_loom_type env dest_t in
        Convert((e', dest_t'), source_t'), dest_t'

  | Ast.Place(p)            ->      (* have to thread the node type t' through *)
      let (p', t') = check_place env p in
        Place(p'), t'

  | Ast.Binop(e1,op,e2)     ->
    begin
      let (e1', t1') as et1' = check_expr env e1 in
      let (e2', t2') as et2' = check_expr env e2 in
      let op' = check_binop op in
        match op' with
      
          Map -> begin
            match (t1', t2') with
              Function([arg_t], ret_t), Array(arr_t, n) ->
                if arg_t <> arr_t then
                  failwith "\n*** map: function arg and array types must match."
                else
                  Binop(et1', op', et2'), Array(ret_t, n)
            | _ -> failwith ("\n*** map operator must have function of one argument"
                            ^ " on left, and array on right")
          end
          
        | Reduce -> begin
            match (t1', t2') with
              Function([arg1_t; arg2_t], ret_t), Array(arr_t, n) ->
                if arg1_t <> arg2_t || arg1_t <> arr_t then
                  failwith "\n*** reduce: function arg and array types must match."
                else
                  Binop(et1', op', et2'), ret_t
            | _ -> failwith ("\n*** reduce operator must have function of two "
                            ^ "arguments on left, and array on right")
          end
          
        | Scanr | Scanl -> begin
            match (t1', t2') with
              Function([arg1_t; arg2_t], ret_t), Array(arr_t, n) ->
                if arg1_t <> arg2_t || arg1_t <> arr_t then
                  failwith "\n*** scan[lr]: function arg and array types must match."
                else
                  Binop(et1', op', et2'), Array(ret_t, n)
            | _ -> failwith ("\n*** reduce operator must have function of two "
                            ^ "arguments on left, and array on right")
          end
          
        | Mult | Div | Plus | Minus | Mod | BitOr | BitAnd | BitXor -> begin
            if t1' <> t2' then
              failwith ("\n*** types for operands of '" ^ Ast.string_of_binop op
                        ^ "' must match exactly.")
            else if t1' = Bool || t1' = Void then
              failwith ("\n*** operands to " ^ Ast.string_of_binop op 
                        ^ "must have numeric type.")
            else
              Binop(et1', op', et2'), t1'
          end

        | Lshift | Rshift -> begin
            if t2' <> Int32 then
              failwith ("\n*** shift operator " ^ Ast.string_of_binop op
                       ^ " expects Int32 on right (try a type conversion?).")
            else if t1' = Bool || t1' = Void then
              failwith ("\n*** operands to " ^ Ast.string_of_binop op 
                        ^ "must have numeric type.")
            else
              Binop(et1', op', et2'), t1'
          end
          
        | And | Or -> begin
            if t1' <> Bool || t2' <> Bool then
              failwith ("\n*** and / or operands must have type Bool.")
            else
              Binop(et1', op', et2'), Bool
          end
          
        | _ -> begin  (* comparison operators *)
            if t1' <> t2' then
              failwith ("\n*** types for operands of '" ^ Ast.string_of_binop op
                       ^ "' must match exactly.")
            else
              Binop(et1', op', et2'), Bool
          end
    end
    
  | Ast.Unop(op,e)          ->
      let (e', t') as et' = check_expr env e in
      let op' = check_unop op in
        if op' = Not then
          if t' <> Bool then
            failwith "\n*** not operand must have type Bool"
          else
            Unop(op', et'), Bool
        else if t' = Bool || t' = Void then
          failwith ("\n*** operand to " ^ Ast.string_of_unop op 
                    ^ "must have numeric type.")
        else
          Unop(op', et'), t'
                  
  | Ast.Call(f,args)        ->
      let args' = List.map (check_expr env) args in
      let _, types' = List.split args' in
      let (fname, formals_t, ret_t) as fd = 
        try
          lookup_func env.scope f
        with Not_found ->
          failwith ("\n*** Undefined function name " ^ f)
      in
        (* check the types of the actual arguments against function signature *)
        if List.for_all (fun (x, y) -> x = y) 
                        (List.combine types' formals_t) then
          Call(fd, args'), ret_t
        else
          failwith ("\n*** arguments to " ^ f ^ "(...) do not match formals.")

  | Ast.NullExpr            ->  NullExpr, Void

and check_place env = 
  let vlookup name =
    try
      lookup_var env.scope name
    with Not_found ->
      failwith ("\n*** No variable or function declaration for identifier " 
                ^ name ^ " in array/record;\nscope trace: \n" 
                ^ string_of_scope env.scope)
  in
  function
  
    Ast.Aref(a,e)           ->
      let (e', t') as et' = check_expr env e in
      let (_, a_t) as vd = vlookup a in
      begin
        match a_t with
          Array(elem_t, _) -> Aref(vd, et'), t'
        | _          -> failwith ("\n*** " ^ a ^ " has non-array type.")
      end

  | Ast.Saref(a,e)          ->
      let (e', t') as et' = check_expr env e in
      let (_, a_t) as vd = vlookup a in
      begin
        match a_t with
          Array(elem_t, _) -> Saref(vd, et'), t'
        | _          -> failwith ("\n*** " ^ a ^ " has non-array type.")
      end

  | Ast.Rref(r,x)           ->
      let (_, r_t) as vd = vlookup r in
      begin
        match r_t with
          Record(fields) ->
            let t' =  try  List.assoc x fields
                      with Not_found -> failwith ("\n*** Unknown field " ^ x)
            in
              Rref(vd, x), t'
        | _           ->  failwith ("\n*** " ^ r ^ " has non-record type.")
      end

  | Ast.Id(s)               ->
      let id =                    (* id has type decl *)
        if List.mem s thread_constants then
          Const(s, Int32)
        else
          try Var(lookup_var env.scope s)
          with Not_found -> 
            try Func(lookup_func env.scope s)
            with Not_found ->
              failwith ("\n*** No variable or function declaration for identifier " 
                        ^ s ^ ";\nscope trace: \n" ^ string_of_scope env.scope)
      in
      begin
        match id with
          Var(name, t)              ->  Id(Var(name,t)), t
        | Func(name, args_t, ret_t) ->  Id(Func(name, args_t, ret_t)),
                                          Function(args_t, ret_t)
        | Const(name, t)            ->  Id(Const(name,t)), t
      end
;;

let rec check_block env stmts =
  (* create a new child scope *)
  let scope' =  { parent = Some(env.scope);
                  vars  = [];
                  funcs = [];
                  types = [] } in
  let env' = { env with scope = scope' } in
  let stmts' = List.map (fun s -> check_stmt env' s) stmts in
    stmts'

and check_stmt env = function
    Ast.Block(stmts) ->
      let stmts' = check_block env stmts in
        Block(stmts')

  | Ast.Vdecl(space, name, t, init) ->
    begin
      if List.exists (fun ((s,_),_) -> s = name) env.scope.vars then
        failwith ("\n*** Attempt to re-declare variable " ^ name)
      else
        let t' = check_loom_type env t in
        let space' = check_memspace space in
          env.scope.vars <-
            ((name, t'), space') :: env.scope.vars;
          match init with
            Some(e) ->  let (_,init_t) as et' = check_expr env e in
                          if init_t <> t' then
                            failwith ("\n*** Variable " ^ name
                                      ^ " initialization has incorrect type.")
                          else
                            Vdecl(space', (name, t'), Some(et'))
          | None    ->  Vdecl(space', (name, t'), None)
    end
     
  | Ast.Tdecl(name, t) ->
      if List.exists (fun (s,_) -> s = name) env.scope.types then
        failwith ("\n*** Attempt to re-declare type " ^ name)
      else
        let t' = check_loom_type env t in
          env.scope.types <- (name, t') :: env.scope.types;
          Tdecl(name,t')  (* FIXME - could probably just be NullStmt *)

  | Ast.Expr(e)   -> Expr(check_expr env e)
  
  | Ast.Return(e) -> 
      let (_, t') as et' = check_expr env e in
        if t' <> env.return_type then
          failwith ("\n*** Return expression (of type "
                    ^ string_of_type t'
                    ^ ") must match function return type "
                    ^ string_of_type env.return_type ^ " in "
                    ^ env.func_name ^ ".")
        else
          Return(et')

  | Ast.Assign(p,e) ->
      let p' = check_place env p in
      let e' = check_expr env e in
        Assign(p', e')

  | Ast.If(e,stmts1,stmts2) ->
      let (_,t') as et' = check_expr env e in
        if t' <> Bool then
          failwith "\n*** Expression in if statement must have type Bool."
        else
          If(et', check_block env stmts1, check_block env stmts2)

  | Ast.For(name,t,e1,e2,stmts) ->
      let t' = check_loom_type env t in
      let (_,t1') as et1' = check_expr env e1 in
      let (_,t2') as et2' = check_expr env e2 in
        if t' <> t1' || t' <> t2' || t' = Bool || t' = Void then
          failwith "\n*** Mismatched or non-numeric types in for loop."
        else
          (* can't use check_block because we have to squeeze in index var *)
          let scope' =  { parent = Some(env.scope);
                          vars  = [((name,t'),Local)];
                          funcs = [];
                          types = [] } in
          let env' = {  func_name = env.func_name  
                      ; return_type = env.return_type 
                      ; loop_level = env.loop_level + 1
                      ; scope = scope' } in
          let stmts' = List.map (fun s -> check_stmt env' s) stmts in
            For((name, t'), et1', et2', stmts')

  | Ast.While(e,stmts) ->
      let (_,t') as et' = check_expr env e in
        if t' <> Bool then
          failwith "\n*** Expression in while statement must have type Bool."
        else
          let env' = { env with loop_level = env.loop_level + 1 } in
            While(et', check_block env' stmts)

  | Ast.Break     ->
      if env.loop_level > 0 then
        Break
      else
        failwith "\n*** Attempt to BREAK outside of loop environment."
        
  | Ast.Continue  ->
      if env.loop_level > 0 then
        Break
      else
        failwith "\n*** Attempt to CONTINUE outside of loop environment."
  
  | Ast.Sync      -> Sync

  | Ast.NullStmt  -> NullStmt
;;

(* Side effect: updates env scope (as in variable declaration). *)
let check_formal env (name, t) =
  let t' = check_loom_type env t in
  let new_var = ((name, t'), Param) in
    env.scope.vars <- new_var :: env.scope.vars;
    (name, t')
;;

let empty_environment name =
  let empty_scope = { parent = None
                    ; vars = []
                    ; types = []
                    ; funcs = [] } in
    { func_name   = name
    ; return_type = Void
    ; loop_level  = 0
    ; scope       = empty_scope }
;;

let extract_fdecl (f : Ast.func) : fdecl =
  let empty_env = empty_environment ""                    in
  let (_, types) = List.split f.Ast.formals               in
  let types' = List.map (check_loom_type empty_env) types in
  let ret_t' = check_loom_type empty_env f.Ast.ret_type   in
    (f.Ast.fname, types', ret_t')
;;

let get_fdecls funcs = List.map extract_fdecl funcs
;;

let check_kernel fdecls k =
  let name      = k.Ast.kname in
  let env       = empty_environment name in
  let kformals' = List.map (check_formal env) k.Ast.kformals in
    env.scope.funcs <- fdecls;
    let kbody'  = check_block env k.Ast.kbody in
      { kname     = name
      ; kformals  = kformals'
      ; kbody     = kbody' }
;;

let check_func (fdecls : fdecl list) (f : Ast.func) =
  let name      = f.Ast.fname in
  let env       = empty_environment name in
  let ret_type' = check_loom_type env f.Ast.ret_type in
  let env'      = { env with return_type = ret_type' } in
  let formals'  = List.map (check_formal env') f.Ast.formals in
    env'.scope.funcs <- fdecls;    
    let body'   = check_block env' f.Ast.body in
      { fname     = f.Ast.fname
      ; formals   = formals'
      ; ret_type  = ret_type'
      ; body      = body' }
;;

let analyze : Ast.program -> program = function
  Some(k), funcs -> 
    let fdecls = get_fdecls funcs in
    let k' = check_kernel fdecls k in
    let funcs' = List.map (check_func fdecls) funcs in
      (k', funcs')
| _              -> failwith "\n*** No kernel function found."
;;

(* Sast pretty printer not yet implemented *)
let to_string _ = "Typechecking finished with no errors.\n"
